// set the dimensions and margins of the graph
const scaleValue = 1.01;

const width = 930 * scaleValue;
const height = 500 * scaleValue;

const mapWidth = width * 0.5;
const plotWidth = width - mapWidth;

const margin = 60;

// Let's create two svg elements and append them to my top-level div
// Here's the first svg, and I'll save the selection to the variable "svg"
const svg = d3
  .select('#my_dataviz')
  .append('svg')
  .attr('width', mapWidth)
  .attr('height', height);

// And the second; the ".append" method returns a selection with just
// the svg in it, so I can manipulate it later (by appending or joining
// data with it)
const svg2 = d3
  .select('#my_dataviz')
  .append('svg')
  .attr('width', plotWidth)
  .attr('height', height);

const tooltip = d3
  .select('#my_dataviz')
  .append('div')
  .attr('id', 'tooltip');

// Here's the main rendering code; whenever I call this function, it
// will draw a choropleth map from the mapData and a scatter plot from
// the countryData
function render(mapData, countryData) {
  // Here's how I'm going to convert lat/long coordinates into pixel values
  // on the screen; I can use whichever projection I like best
  var projection = d3
    .geoEqualEarth()
    .fitSize([mapWidth, height], mapData);
  // And here's the part that converts a sequence of points around a
  // polygon into a valid "d" attribute for a <path> tag
  var path = d3.geoPath(projection);

  // Here are the names of the attributes I'll pull out of my data
  // objects for my various visual channels
  // It's useful to declare them once, since I will need to refer to them
  // in several places (e.g. setting visual properties, axes, and labels)
  const mapColorAttr = 'country';

  const plotColorAttr = 'country';
  const xAttr = 'max_life_expectancy';
  const yAttr = 'min_weight';
  // const rAttr = 'Population';
  const rAttr = (d) =>
    (d.name.length * d[xAttr]) / d[yAttr];

  // Here's a snippet of code that I ended up using in a couple places,
  // so I've refactored it out as a function
  //     data: an iterable (typically Array) of all data points
  //     attr: the name of the attribute you want the extent of
  //     returns: [min, max] of the attr values in data
  // Note: min and max only really makes sense for numeric types, so I've
  // included a typecast (the "+") - this will keep me from accidentally
  // getting tripped up when my attribute is really strings representing
  // numbers
  function getDataExtent(data, attr) {
    return d3.extent(
      // For my data, missing values show up as empty strings, which
      // JavaScript then casts to 0. I will filter out all of these to make
      // sure I get the true extent of the non-missing data
      d3.map(data, (r) => +r[attr]).filter((r) => r > 0),
    );
  }

  // Here's a scale that converts attribute values into a color that is
  // on a linear scale between white and blue
  var mapColorScale = d3
    // The type of scale: linear, log, band, ordinal, etc.
    .scaleLinear()
    // The range of possible input values:
    .domain(
      // getDataExtent(countryData.values(), 'count'), //mapColorAttr),
      getDataExtent(countryData.values(), 'count'), //mapColorAttr),
    )
    // And the range of possible output values:
    .range(['black', 'red']);
  // console.log(
  //   'countryData.values(): ',
  //   countryData.values(),
  // );
  // for (const item of countryData.values()) {
  //   console.log('iter item: ', item['name']);
  // }
  console.log(
    'getDataExtent(countryData.values(), count): ',
    getDataExtent(countryData.values(), 'count'),
  );

  // Here are the rest of the scaling functions (for the scatter plot)
  // We'll need to handle the x/y positions, radius, and color

  // Boring linear scale for x
  const x = d3
    .scaleLinear()
    .domain(getDataExtent(countryData.values(), xAttr))
    .range([margin, plotWidth - margin]);

  // Boring linear scale for y
  const y = d3
    .scaleLinear()
    .domain(getDataExtent(countryData.values(), yAttr))
    // Have you noticed that the range for the y axis is backwards?
    // It maps the lowest input to the highest pixel value, and the highest
    // input to 0 pixels. This is because the origin (0,0) for graphics is
    // in the top left, instead of the bottom left like mathematicians might
    // expect
    .range([height - margin, margin]);

  const plotColorScale = d3.scaleOrdinal(
    d3.schemeCategory10,
  );

  // const rScale = d3
  //   .scaleLinear()
  //   .domain(getDataExtent(countryData.values(), rAttr))
  //   .range([3, 10]);
  const rScale = d3
    .scaleLinear()
    .domain(getDataExtent(countryData.values(), rAttr))
    .range([3, 10]);

  function getByValue(map, searchValue) {
    for (let [key, value] of map.entries()) {
      // console.log(
      //   'value[country]: ',
      //   value['country'],
      //   searchValue,
      // );
      if (value['country'] === searchValue) return key;
    }
  }
  // Since we need to do a little error checking here, lets write a
  // nice function so that our attribute setting looks cleaner
  function getCountryColor(d) {
    // First, try to get the data for a given country
    // let people = new Map();
    // people.set('1', 'jhon');
    // people.set('2', 'abdo');
    // people.set('3', 'jhon');

    // console.log(getByValue(people, 'jhon'));
    // console.log(getByValue(people, 'abdo'));

    const countryName = getByValue(
      countryData,
      d.properties.name,
    );
    console.log('countryName: ', countryName);
    // let entry = countryData.get(countryName); //d.properties.name);
    let entry = countryData; //d.properties.name);

    // console.log(
    //   'getCountryColor country 1: ',
    //   // d.properties.name,
    //   // d.properties,
    //   d,
    //   d3.map(countryData.values(), (r) => r['country']),
    // );
    // And check if we actually retrieved a value; if not, we'll only
    // get "undefined" for missing data and can use a default 'grey' color
    return entry == undefined
      ? 'grey'
      : mapColorScale(entry['country']); //mapColorAttr]);
  }

  // This one is way more straightforward, since I assume I always have
  // complete data for each point on the scatter plot
  function getPointColor(d) {
    return plotColorScale(d[plotColorAttr]);
  }
  //=============== MAP here is where they are adding path elements...=======================================
  // Ok, down to brass tacks: let's add <path> elements to our first svg
  // that encode the borders of each country. I'll save this selection as
  // a variable, so that I can modify the countries later if I need to
  const countryMarks = svg
    // I like to append a <g> element that will hold all my paths, just so
    // I have the option of easily adjusting or moving the entire group
    .append('g')
    // Since the svg is originally empty (I just created it, after all), any
    // selection here will be empty; its only purpose is to give me a place
    // to join my data
    .selectAll('path')
    // Once I have the (empty) selection, I will join all of my data points to
    // it. I know that the structure of geojson is for there to be a "features"
    // attribute that contains the Array of all the geo objects, and I want to
    // bind each geo object to its own path
    .data(mapData.features)
    .join('path')
    // Ok, now that I have a shiny new <path> element with a single country's
    // geo data bound to it, I'm ready to start setting them up.
    // First, the fill color: I'll use the function I wrote above that converts
    // a country name string into a scaled color value
    .attr('fill', getCountryColor)
    // Next, I'll set the "id" attribute, which will give me the ability to
    // select linked items that have the same id
    .attr('id', (d) => d.properties.name)
    .classed('poly', true) // don't use this
    // .attr('class', (d) => `${d.country}`)

    // Finally, lets use the default geojson setup of setting the 'd' attribute
    // using our path generator function. This is sort of magic that works under
    // the hood
    .attr('d', path);

  // x axis
  svg2
    .append('g')
    .attr('transform', `translate(0,${height - margin})`) // This controls the vertical position of the Axis
    .call(d3.axisBottom(x));

  svg2
    .append('text')
    .text(xAttr)
    .attr('x', plotWidth / 2)
    .attr('y', height - margin / 2)
    .style('text-anchor', 'middle');

  // y axis
  svg2
    .append('g')
    .attr('transform', `translate(${margin},0)`) // This controls the vertical position of the Axis
    .call(d3.axisRight(y));

  svg2
    .append('text')
    .text(yAttr)
    .attr('transform', 'rotate(-90)')
    .attr('x', -height / 2)
    .attr('y', margin / 2);

  // Now we need to add points to our scatter plot; again I will save the
  // selection as a variable so I can modify it more later
  const scatterMarks = svg2
    .append('g')
    .selectAll('circle')
    .data(countryData.values())
    .join('circle')
    .attr('cx', (d) => x(d[xAttr]))
    .attr('cy', (d) => y(d[yAttr]))
    // .attr('r', (d) => rScale(d[rAttr]))
    .attr('r', rAttr) //d.count) // rScale(d['name'])) //'10')

    .attr('fill', getPointColor)
    .attr('id', (d) => d.name)
    // .attr('class', (d) => `dot ${d.Continent}`)
    .attr('class', (d) => `${d.country}`)

    .style('opacity', 0.4);
  console.log(countryData.values());

  // console.log(countryData.values().xAttr);
  // Okay! It's later now: let's hook up the mouse events for our countries
  // and points in the scatter plot

  // Planning ahead, the first thing I'm going to want to do is highlight
  // matching marks in the linked plots, and I know they're linked by their
  // "id" attribute matching, and differentiated by having different classes
  // ("poly" for the map marks, and "dot" for the scatterplot marks)
  function highlightMarks(id) {
    // Select everything with this id (polygon and dot), and set the fill
    // color for both
    d3.selectAll(`#${'class'}.poly`) //id}.poly`)
      .attr('fill', 'purple')
      .style('opacity', 1);
    // Select just the dot with the right id, and set its radius
    d3.selectAll(`#${'class'}.dot`) //id}.dot`)
      .transition()
      .duration(750)
      .ease(d3.easeBounce)
      .attr('r', 20)
      .style('opacity', 1)
      .style('stroke', 'black');
  }

  // Finally, I can use my scaling functions for color and radius to reset
  // back to normal when I mouse out
  function resetMarks(id) {
    d3.selectAll(`#${'class'}.dot`) // id}.dot`)
      .transition()
      .attr('fill', getPointColor)
      .attr('r', (d) => rScale(d[rAttr]))
      .style('opacity', 0.4)
      .style('stroke', 'none');
    d3.selectAll(`#${id}.poly`)
      .transition()
      .duration(600)
      .attr('fill', getCountryColor);
  }

  function highlightContinent(continent) {
    d3.selectAll(`.${continent}.dot`)
      .style('stroke', 'black')
      .style('opacity', 1);
  }

  function unHighlightContinent(continent) {
    d3.selectAll(`.${continent}.dot`)
      .style('stroke', 'none')
      .style('opacity', 0.4);
  }

  function popUpTooltip(event, data) {
    let country = countryData.get(data.properties.name);
    console.log(name, countryData.has(name));
    tooltip.style('opacity', 1);
    tooltip.html(
      // `${country.Country_orig}<br/>population ${country.Population_orig}`,
      `${country.Country_orig}<br/>population ${country.Population_orig}`,
    );
  }

  function adjustTooltip(event) {
    tooltip
      .style('left', event.x - 80 + 'px')
      .style('top', event.y - 80 + 'px');
  }

  function addCentroid(continent) {
    console.log(continent);
    let validCountries = [...countryData.values()].filter(
      (d) => d.Continent == continent,
    );
    let xVal = d3.mean(
      d3.map(validCountries, (d) => d[xAttr]),
    );
    let yVal = d3.mean(
      d3.map(validCountries, (d) => d[yAttr]),
    );
    svg2
      .append('circle')
      .attr('id', 'mean')
      .attr('cx', x(xVal))
      .attr('cy', height - margin)
      .attr('fill', 'black')
      .attr('r', 5);
    svg2
      .append('circle')
      .attr('id', 'mean')
      .attr('cx', margin)
      .attr('cy', y(yVal))
      .attr('fill', 'black')
      .attr('r', 5);
  }

  function removeCentroid() {
    d3.selectAll('#mean').remove();
  }

  // With these behaviors captured in functions, we just need to hook them up
  // as event listeners on our mark selections!

  // countryMarks // countryMarks
  //   .on('mouseover', function (event, data) {
  //     highlightMarks(this.id);
  //     popUpTooltip(event, data);
  //   })
  //   .on('mousemove', function (event, data) {
  //     adjustTooltip(event);
  //   })
  //   .on('mouseleave', function (event, data) {
  //     resetMarks(this.id);
  //     tooltip.style('opacity', 0);
  //   });

  // scatterMarks
  //   .on('mouseover', function (event, data) {
  //     highlightContinent(data.name);
  //     highlightMarks(this.id);
  //     addCentroid(data.name);
  //   })
  //   .on('mouseleave', function (event, data) {
  //     unHighlightContinent(data.Continent);
  //     resetMarks(this.id);
  //     removeCentroid();
  //   });
  //   .on('mouseover', function (event, data) {
  //     highlightMarks(this.id);
  //     popUpTooltip(event, data);
  //   })
  //   .on('mousemove', function (event, data) {
  //     adjustTooltip(event);
  //   })
  //   .on('mouseleave', function (event, data) {
  //     resetMarks(this.id);
  //     tooltip.style('opacity', 0);
  //   });

  // scatterMarks
  //   .on('mouseover', function (event, data) {
  //     highlightContinent(data.name);
  //     highlightMarks(this.id);
  //     addCentroid(data.name);
  //   })
  //   .on('mouseleave', function (event, data) {
  //     unHighlightContinent(data.Continent);
  //     resetMarks(this.id);
  //     removeCentroid();
  //   });
}

//Read the data
const mapUrl =
  'https://raw.githubusercontent.com/holtzy/D3-graph-gallery/master/DATA/world.geojson';
const catsUrl = 'new_cat_breeds_count.csv';
// const countryStatsUrl =
//   'https://raw.githubusercontent.com/bepnye/CP341/main/world-data-2023.csv';
const continentUrl =
  'https://raw.githubusercontent.com/samayo/country-json/master/src/country-by-continent.json';

// methods for parsing data:
function normalizeCountryName(s) {
  return s.toLowerCase();
}

function parseNumStr(s) {
  return +s.replaceAll(',', '').replaceAll('$$', '');
}

function parseMultipleCountries(s) {
  return s.replaceAll('&', ' ');
}

function findAnd(s) {
  s.includes('&') ? console.log(s) : console.log('false');
}
// aka main ()
Promise.all([
  d3.json(mapUrl),
  d3.csv(catsUrl),
  // d3.csv(countryStatsUrl),
  d3.json(continentUrl),
]).then(([mapData, catsData, continentData]) => {
  var countryData = new Map();
  // console.log(catsData);
  // console.log(continentData);

  // processing data...
  var countryToMapNames = new Map([
    ['United States', 'USA'],
    ['Tanzania', 'United Republic of Tanzania'],
    ['United Kingdom', 'England'],
    ['Republic of Ireland', 'Ireland'],
    // [
    //   'Great Britain',
    //   'England&Northern_Ireland&Wales&Scotland',
    // ],
    // [
    //   'Southeast Asia',
    //   'Cambodia&Laos&Vietnam&Myanmar&Thailand&Malaysia&Brunei&East_Timor&Indonesia&Philippines&Singapore',
    // ],
  ]);

  // process catsData ...
  catsData.forEach((row) => {
    let country = row.country;
    console.log(country);

    if (countryToMapNames.has(country)) {
      // console.log('found a country');
      country = countryToMapNames.get(country);
    }
    // console.log('After country to Map names', country);
    // if (countryToMapNames.has(row.Country)) {
    //   row.Country = countryToMapNames.get(row.Country);
    // }
    // Cleaning specific columns of my data
    // Each row is an object with attributes corresponding to the columns,
    // so I am going to overwrite the values for some of the attributes
    // to be the correct type/format

    // for cat data:
    row.Country_orig = country;
    row.country = normalizeCountryName(country);
    // row.country = normalizeCountryName(country);
    //https://www.freecodecamp.org/news/how-to-use-javascript-collections-map-and-set/

    // can you dectect & character
    // findAnd(country);
    // console.log('after normalizing', country);

    // for World data:
    // row.Country_orig = row.Country;
    // row.Country = normalizeCountryName(row.Country);
    // row.Population_orig = row.Population;
    // row.Population = parseNumStr(row.Population);
    // row.Urban_population = parseNumStr(
    //   row.Urban_population,
    // );
    // row.Urban_rate = row.Urban_population / row.Population;
    // row.GDP_per_capita =
    //   parseNumStr(row.GDP) / row.Population;

    // Since I'm going to be pulling in an extra dataset with Continent data,
    // I'll initialize a default "unknown" value that I will (hopefully!)
    // overwrite with the actual continent name. This is useful since now if
    // I'm missing data for any country, it will have a valid attribute when I
    // select the .Continent, saving me from anoying "undefined" bugs
    row.Continent = 'Unknown';
    // countryData.set(country, row);
    countryData.set(row.name, row);
  });

  console.log('post processed', catsData);
  console.log('mapData: ', mapData);
  mapData.features.forEach((f) => {
    // console.log(f);
    f.properties.name = normalizeCountryName(
      f.properties.name,
    );
  });
  // assigns a continent to a given country
  // continentData.forEach((row) => {
  //   let name = normalizeCountryName(row.country);
  //   if (countryData.has(name)) {
  //     console.log('^&: ', countryData.get(name));
  //     countryData.get(name).Continent =
  //       normalizeCountryName(row.continent);
  //   } else {
  //     // console.log('No match found for ', name);
  //   }
  // });
  console.log('countryData', countryData);
  // console.log(continentData);
  render(mapData, catsData);
});
